// encapsulate original function
function encapsulate (object, variable, fn)
{
  object[variable + "Orig"] = object[variable];
  object[variable] = fn;
}

function assert(condition, message) {
    if (!condition) {
        message = message || "Assertion failed";
        if (typeof Error !== "undefined") {
            throw new Error(message);
        }
        throw message; // Fallback
    }
}

/**
* @class Zeměpisná poloha - dvojice souřadnic
* @group map
*/
SMap.Coords = JAK.ClassMaker.makeClass({
NAME:"SMap.Coords",
VERSION:"1.0"
});

/**
* Získá latitude
* @author N
* @returns {int}
*/
SMap.Coords.prototype.lat = function ()
{
  return this._lat;
}

/**
* Získá longitude
* @author N
* @returns {int}
*/
SMap.Coords.prototype.lon = function ()
{
  return this._lon;
}

/**
* Konstruktor je lepší nepoužívat, namísto toho vyrábět souřadnice pomocí továrních metod
* @author Seznam, upraveno by N
* @param {int} lat Latitude
* @param {int} lon Longitude
*/
SMap.Coords.prototype.$constructor = function(lat, lon)
{
  this._lat = lat || 0;
  this._lon = lon || 0;
  
  // normalize lon to -180..180
  this._lon = ((this._lon + 3780) % 360) - 180;
  
  // cut off lat to -90..90
  if (this.lat > 90)
    this.lat = 90;
  else if (this.lat < -90)
    this.lat = -90;
   
//   this.x = this._lat;
//   this.y = this._lon;
  
  //assert(lat >= -1000 && lat <= 1000 && lon >= -5000 && lon <= 5000, lat + "\n" + lon);
}

/**
* Tovární metoda - výroba souřadnic z PP
* @author Seznam
* @param {int} PPx Xová souřadnice
* @param {int} PPy Yová souřadnice
* @returns {SMap.Coords}
*/
SMap.Coords.fromPP = function(PPx, PPy)
{
  var x = PPx / 32 + (-3700000);
  var y = PPy / 32 + (1300000);
  
  return SMap.Coords.fromUTM33(x, y);  
}

/**
* Převede souřadnice do PP
* @returns {float[]} Pole [PPx, PPy]
*/
SMap.Coords.prototype.toPP = function()
{
  var utm = this.toUTM(33);
  var x = (utm.x - (-3700000)) * 32;
  var y = (utm.y - (1300000)) * 32;
  return [Math.round(x), Math.round(y)];
}

/**
* Tovární metoda - výroba souřadnic z UTM33
* @author: N
* @param {int} x Xová souřadnice
* @param {int} y Yová souřadnice
* @returns {SMap.Coords}
*/
SMap.Coords.fromUTM33 = function(x, y)
{
  var ret = SMap.Coords.fromUTM(x, y, 33, false);
  
  // pokud dostaneme pro UTM33 nesmyslné hodnoty, tak vrátíme nějaké rozumné souřadnice
  // -- je třeba někdy implementovat, aby mapa taková místa vůbec nepožadovala
  // ---> vypnuto, v US při přepnutí google maps na mapy.cz ophoto crash
  //if (ret.lat() >= 90 || ret.lat() <= -90 || ret.lon() < -180 || ret.lat() > 180)
  //  return new SMap.Coords(50, 15);

  return ret;
}

/**
* Převede souřadnice do UTM33
* @returns {float[]} Pole [x, y]
*/
SMap.Coords.prototype.toUTM33 = function()
{
  var utm = this.toUTM(33);
  return [utm.x, utm.y];
}

/**
* Tovární metoda - výroba souřadnic z WGS 84
* @author N
* @param {float} lat Longitude (zeměpisná délka) ve stupních
* @param {float} lon Latitude (zeměpisná šířka) ve stupních
* @returns {SMap.Coords}
*/
SMap.Coords.fromWGS = function(lat, lon)
{
  //assert(typeof lat != "string" && typeof lon != "string");
  
  lat = parseFloat(lat);
  lon = parseFloat(lon);

  return new SMap.Coords(lat, lon);
}

/**
* Tovární metoda - výroba souřadnic z WGS 84 (!! nepoužívat !!)
* @author Seznam, zkráceno by N
* @param {float} lat Longitude (zeměpisná délka) ve stupních
* @param {float} lon Latitude (zeměpisná šířka) ve stupních
* @returns {SMap.Coords}
*/
SMap.Coords.fromWGS84 = function(lonD, latD)
{
  if (arguments.length == 1)
    return SMap.Coords.fromString(arguments[0]) || assert(false);
  else if (typeof(latD) == "string" || typeof(lonD) == "string")
    return SMap.Coords.fromString(latD + " " + lonD) || assert(false);

  return SMap.Coords.fromWGS(latD, lonD);
}

/**
* Převede souřadnice do WGS 84 (!! nepoužívat !!)
* @author Seznam, upraveno by N
* @param {int} [format] Formát pro převod na řetězec: čím větší číslo, tím detailnější je řetězec (0 = stupně, 1 = minuty, 2 = vteřiny). Při nezadání bude vrácena dvojice floatů.
* @returns {float[] || string[]} Pole [longitude, latitude]
*/
SMap.Coords.prototype.toWGS84 = function(format)
{
  var wgs = [this.lon(), this.lat()];
 	return (typeof(format) == "number") ? this.format("WGS", (format == 0) ? "deg" : ((format == 1) ? "min" : "sec")) : wgs;   
}

/**
* Tovární metoda - výroba souřadnic z UTM
* @author Chuck Taylor (http://home.hiwaay.net/~taylorc/toolbox/geography/geoutm.html)
* @param {int} x easting
* @param {int} y northing
* @param {int} zone zóna
* @param {boolean} [southhemi=false] leží bod na jižní polokouli?
* @returns {SMap.Coords}
*/
SMap.Coords.fromUTM = function(x, y, zone, southhemi)
{
  /* Ellipsoid model constants (actual values here are for WGS84) */
  var sm_a = 6378137.0;
  var sm_b = 6356752.314;
  var sm_EccSquared = 6.69437999013e-03;
  var UTMScaleFactor = 0.9996;
  
  var cmeridian;
  	
  x -= 500000.0;
  x /= UTMScaleFactor;
  	
  /* If in southern hemisphere, adjust y accordingly. */
  if (southhemi)
    y -= 10000000.0;
  		
  y /= UTMScaleFactor;
  
  var rad = Math.PI / 180;
  cmeridian = (-183.0 + (zone * 6.0)) * rad;
  
  var latlon = new Array(2);
  var lambda0 = cmeridian, philambda = latlon;
  
  var phif, Nf, Nfpow, nuf2, ep2, tf, tf2, tf4, cf;
  var x1frac, x2frac, x3frac, x4frac, x5frac, x6frac, x7frac, x8frac;
  var x2poly, x3poly, x4poly, x5poly, x6poly, x7poly, x8poly;

  var y_ = y, alpha_, beta_, gamma_, delta_, epsilon_, n;
  var result;
  
  n = (sm_a - sm_b) / (sm_a + sm_b);
  alpha_ = ((sm_a + sm_b) / 2.0) * (1 + (Math.pow (n, 2.0) / 4) + (Math.pow (n, 4.0) / 64));
  y_ = y / alpha_;
  beta_ = (3.0 * n / 2.0) + (-27.0 * Math.pow (n, 3.0) / 32.0) + (269.0 * Math.pow (n, 5.0) / 512.0);
  gamma_ = (21.0 * Math.pow (n, 2.0) / 16.0) + (-55.0 * Math.pow (n, 4.0) / 32.0);
  delta_ = (151.0 * Math.pow (n, 3.0) / 96.0) + (-417.0 * Math.pow (n, 5.0) / 128.0);
  epsilon_ = (1097.0 * Math.pow (n, 4.0) / 512.0);

  phif = y_ + (beta_ * Math.sin (2.0 * y_))
      + (gamma_ * Math.sin (4.0 * y_))
      + (delta_ * Math.sin (6.0 * y_))
      + (epsilon_ * Math.sin (8.0 * y_));

  ep2 = (Math.pow (sm_a, 2.0) - Math.pow (sm_b, 2.0)) / Math.pow (sm_b, 2.0);
  cf = Math.cos (phif);
  nuf2 = ep2 * Math.pow (cf, 2.0);
  Nf = Math.pow (sm_a, 2.0) / (sm_b * Math.sqrt (1 + nuf2));
  Nfpow = Nf;
  tf = Math.tan (phif);
  tf2 = tf * tf;
  tf4 = tf2 * tf2;
  
  /* Precalculate fractional coefficients for x**n in the equations
     below to simplify the expressions for latitude and longitude. */
  x1frac = 1.0 / (Nfpow * cf);
  Nfpow *= Nf;   /* now equals Nf**2) */
  x2frac = tf / (2.0 * Nfpow);
  Nfpow *= Nf;   /* now equals Nf**3) */
  x3frac = 1.0 / (6.0 * Nfpow * cf);
  Nfpow *= Nf;   /* now equals Nf**4) */
  x4frac = tf / (24.0 * Nfpow);
  Nfpow *= Nf;   /* now equals Nf**5) */
  x5frac = 1.0 / (120.0 * Nfpow * cf);
  Nfpow *= Nf;   /* now equals Nf**6) */
  x6frac = tf / (720.0 * Nfpow);
  Nfpow *= Nf;   /* now equals Nf**7) */
  x7frac = 1.0 / (5040.0 * Nfpow * cf);
  Nfpow *= Nf;   /* now equals Nf**8) */
  x8frac = tf / (40320.0 * Nfpow);
  
  /* Precalculate polynomial coefficients for x**n.
     -- x**1 does not have a polynomial coefficient. */
  x2poly = -1.0 - nuf2;
  x3poly = -1.0 - 2 * tf2 - nuf2;
  x4poly = 5.0 + 3.0 * tf2 + 6.0 * nuf2 - 6.0 * tf2 * nuf2 - 3.0 * (nuf2 *nuf2) - 9.0 * tf2 * (nuf2 * nuf2);
  x5poly = 5.0 + 28.0 * tf2 + 24.0 * tf4 + 6.0 * nuf2 + 8.0 * tf2 * nuf2;
  x6poly = -61.0 - 90.0 * tf2 - 45.0 * tf4 - 107.0 * nuf2 + 162.0 * tf2 * nuf2;
  x7poly = -61.0 - 662.0 * tf2 - 1320.0 * tf4 - 720.0 * (tf4 * tf2);
  x8poly = 1385.0 + 3633.0 * tf2 + 4095.0 * tf4 + 1575 * (tf4 * tf2);
  	
  /* Calculate latitude */
  philambda[0] = phif + x2frac * x2poly * (x * x)
  	+ x4frac * x4poly * Math.pow (x, 4.0)
  	+ x6frac * x6poly * Math.pow (x, 6.0)
  	+ x8frac * x8poly * Math.pow (x, 8.0);
  	
  /* Calculate longitude */
  philambda[1] = lambda0 + x1frac * x
  	+ x3frac * x3poly * Math.pow (x, 3.0)
  	+ x5frac * x5poly * Math.pow (x, 5.0)
  	+ x7frac * x7poly * Math.pow (x, 7.0);

  return (new SMap.Coords.fromWGS(latlon[0] / rad, latlon[1] / rad));
}

/**
* Převede souřadnice do UTM
* @author Chuck Taylor (http://home.hiwaay.net/~taylorc/toolbox/geography/geoutm.html)
* @param {int} [zone=] zóna
* @returns {Object} utmCoords UTM souřadnice
* @returns {number} utmCoords.x easting
* @returns {number} utmCoords.y northing
* @returns {number} utmCoords.zone zóna
*/
SMap.Coords.prototype.toUTM = function(zone)
{
  var zone = zone || Math.floor ((this.lon() + 180.0) / 6) + 1;
  
  var rad = Math.PI / 180;
  var lat = this.lat() * rad;
  var lon = this.lon() * rad;

  /* Ellipsoid model constants (actual values here are for WGS84) */
  var sm_a = 6378137.0;
  var sm_b = 6356752.314;
  var sm_EccSquared = 6.69437999013e-03;
  var UTMScaleFactor = 0.9996;

  var xy = new Array(2);

  var phi = lat, lambda = lon, lambda0 = (-183.0 + (zone * 6.0)) * rad;
  
  var N, nu2, ep2, t, t2, l;
  var l3coef, l4coef, l5coef, l6coef, l7coef, l8coef;
  var tmp;

  ep2 = (Math.pow (sm_a, 2.0) - Math.pow (sm_b, 2.0)) / Math.pow (sm_b, 2.0);
  nu2 = ep2 * Math.pow (Math.cos (phi), 2.0);
  N = Math.pow (sm_a, 2.0) / (sm_b * Math.sqrt (1 + nu2));
  t = Math.tan (phi);
  t2 = t * t;
  tmp = (t2 * t2 * t2) - Math.pow (t, 6.0);

  l = lambda - lambda0;

  l3coef = 1.0 - t2 + nu2;
  l4coef = 5.0 - t2 + 9 * nu2 + 4.0 * (nu2 * nu2);
  l5coef = 5.0 - 18.0 * t2 + (t2 * t2) + 14.0 * nu2 - 58.0 * t2 * nu2;
  l6coef = 61.0 - 58.0 * t2 + (t2 * t2) + 270.0 * nu2 - 330.0 * t2 * nu2;
  l7coef = 61.0 - 479.0 * t2 + 179.0 * (t2 * t2) - (t2 * t2 * t2);
  l8coef = 1385.0 - 3111.0 * t2 + 543.0 * (t2 * t2) - (t2 * t2 * t2);

  /* Calculate easting (x) */
  xy[0] = N * Math.cos (phi) * l
      + (N / 6.0 * Math.pow (Math.cos (phi), 3.0) * l3coef * Math.pow (l, 3.0))
      + (N / 120.0 * Math.pow (Math.cos (phi), 5.0) * l5coef * Math.pow (l, 5.0))
      + (N / 5040.0 * Math.pow (Math.cos (phi), 7.0) * l7coef * Math.pow (l, 7.0));

  /* Calculate northing (y) */
  var alpha, beta, gamma, delta, epsilon, n;

  n = (sm_a - sm_b) / (sm_a + sm_b);
  alpha = ((sm_a + sm_b) / 2.0) * (1.0 + (Math.pow (n, 2.0) / 4.0) + (Math.pow (n, 4.0) / 64.0));
  beta = (-3.0 * n / 2.0) + (9.0 * Math.pow (n, 3.0) / 16.0) + (-3.0 * Math.pow (n, 5.0) / 32.0);
  gamma = (15.0 * Math.pow (n, 2.0) / 16.0) + (-15.0 * Math.pow (n, 4.0) / 32.0);
  delta = (-35.0 * Math.pow (n, 3.0) / 48.0) + (105.0 * Math.pow (n, 5.0) / 256.0);
  epsilon = (315.0 * Math.pow (n, 4.0) / 512.0);
  var ArcLengthOfMeridian = alpha * (phi + (beta * Math.sin (2.0 * phi)) 
    + (gamma * Math.sin (4.0 * phi)) + (delta * Math.sin (6.0 * phi)) + (epsilon * Math.sin (8.0 * phi)));

  xy[1] = ArcLengthOfMeridian
      + (t / 2.0 * N * Math.pow (Math.cos (phi), 2.0) * Math.pow (l, 2.0))
      + (t / 24.0 * N * Math.pow (Math.cos (phi), 4.0) * l4coef * Math.pow (l, 4.0))
      + (t / 720.0 * N * Math.pow (Math.cos (phi), 6.0) * l6coef * Math.pow (l, 6.0))
      + (t / 40320.0 * N * Math.pow (Math.cos (phi), 8.0) * l8coef * Math.pow (l, 8.0));
     
  /* Adjust easting and northing for UTM system. */
  xy[0] = xy[0] * UTMScaleFactor + 500000.0;
  xy[1] = xy[1] * UTMScaleFactor;

  if (xy[1] < 0.0)
      xy[1] = xy[1] + 10000000.0;

  return {x: xy[0], y: xy[1], zone: zone};
}

/**
* Převede souřadnice do S-JTSK (EPSG 102067)
* @author N (převedeno z už neznámého zdroje)
* @returns {Object} sjtskCoords SJTSK souřadnice
* @returns {number} sjtskCoords.x X
* @returns {number} sjtskCoords.y Y
*/
SMap.Coords.prototype.toSJTSK = function()
{
  var lat = this.lat();
  var lon = this.lon();
  
  if (lat < 40 || lat > 60 || lon < 5 || lon > 25)
    return {x: 0, y: 0};

  function sqr(x) {return x * x;}

  var d2r = Math.PI / 180;
  var a = 6378137.0;
  var f1 = 298.257223563;
  var dx = -570.69;
  var dy = -85.69;
  var dz = -462.84;
  var wx = 4.99821/3600 * Math.PI/180;
  var wy = 1.58676/3600 * Math.PI/180;
  var wz = 5.2611/3600 * Math.PI/180;
  var m  = -3.543e-6;

  var B = lat * d2r;
  var L = lon * d2r;
  var H = 0;
  
  var e2 = 1 - sqr(1 - 1 / f1);
  var rho = a / Math.sqrt(1 - e2 * sqr(Math.sin(B)));
  var x1 = (rho + H) * Math.cos(B) * Math.cos(L);
  var y1 = (rho + H) * Math.cos(B) * Math.sin(L);
  var z1 = ((1 - e2) * rho + H) * Math.sin(B);
  
  var x2 = dx + (1 + m) * (x1 + wz * y1 - wy * z1);
  var y2 = dy + (1 + m) * ( - wz * x1 + y1 + wx * z1);
  var z2 = dz + (1 + m) * (wy * x1 - wx * y1 + z1);
  
  a = 6377397.15508;
  f1 = 299.152812853;
  var ab = f1 / (f1 - 1);
  var p = Math.sqrt(sqr(x2) + sqr(y2));
  e2 = 1 - sqr(1 - 1 / f1);
  var th = Math.atan(z2 * ab / p);
  var st = Math.sin(th);
  var ct = Math.cos(th);
  var t = (z2 + e2 * ab * a * (st * st * st)) / (p - e2 * a * (ct * ct * ct));
  
  B = Math.atan(t);
  H = Math.sqrt(1 + t * t) * (p - a / Math.sqrt(1 + (1 - e2) * t * t));
  L = 2 * Math.atan(y2 / (p + x2));
  
  a = 6377397.15508;
  var e = 0.081696831215303;
  var n = 0.97992470462083;
  var rho0 = 12310230.12797036;
  var sinUQ = 0.863499969506341;
  var cosUQ = 0.504348889819882;
  var sinVQ = 0.420215144586493;
  var cosVQ = 0.907424504992097;
  var alpha = 1.000597498371542;
  var k2 = 1.00685001861538;
  
  var sinB = Math.sin(B);
  t = (1 - e * sinB) / (1 + e * sinB);
  t = sqr(1 + sinB) / (1 - sqr(sinB)) * Math.exp(e * Math.log(t));
  t = k2 * Math.exp(alpha * Math.log(t));
  
  var sinU = (t - 1) / (t + 1);
  var cosU = Math.sqrt(1 - sinU * sinU);
  var V = alpha * L;
  var sinV = Math.sin(V);
  var cosV = Math.cos(V);
  var cosDV = cosVQ * cosV + sinVQ * sinV;
  var sinDV = sinVQ * cosV - cosVQ * sinV;
  var sinS = sinUQ * sinU + cosUQ * cosU * cosDV;
  var cosS = Math.sqrt(1 - sinS * sinS);
  var sinD = sinDV * cosU / cosS;
  var cosD = Math.sqrt(1 - sinD * sinD);
  
  var eps = n * Math.atan(sinD / cosD);
  rho = rho0 * Math.exp( - n * Math.log((1 + sinS) / cosS));
  
  var CX = rho * Math.sin(eps);
  var CY = rho * Math.cos(eps);

  return {x: CX, y: CY};    
}

/**
* Tovární metoda - výroba souřadnic z S-JTSK (EPSG 102067)
* @author N (převedeno z neznámého zdroje)
* @param {int} x X
* @param {int} y Y
* @returns {SMap.Coords}
*/
SMap.Coords.fromSJTSK = function(y, x)
{
  /* vstupní údaje */
  var X = x;
  var Y = y;
  var H = 45;
  
  /* Vypocet zemepisnych souradnic z rovinnych souradnic */ 
  var a = 6377397.15508;
  var e = 0.081696831215303;
  var n = 0.97992470462083;
  var konst_u_ro = 12310230.12797036;
  var sinUQ = 0.863499969506341;
  var cosUQ = 0.504348889819882;
  var sinVQ = 0.420215144586493;
  var cosVQ = 0.907424504992097;
  var alfa = 1.000597498371542;
  var k = 1.003419163966575;
  var ro = Math.sqrt(X * X + Y * Y);
  var epsilon = 2 * Math.atan(Y / (ro + X));
  var D = epsilon / n;
  var S = 2 * Math.atan(Math.exp(1 / n * Math.log(konst_u_ro / ro))) - Math.PI / 2;
  var sinS = Math.sin(S);
  var cosS = Math.cos(S);
  var sinU = sinUQ * sinS - cosUQ * cosS * Math.cos(D);
  var cosU = Math.sqrt(1 - sinU * sinU);
  var sinDV = Math.sin(D) * cosS / cosU;
  var cosDV = Math.sqrt(1 - sinDV * sinDV);
  var sinV = sinVQ * cosDV - cosVQ * sinDV;
  var cosV = cosVQ * cosDV + sinVQ * sinDV;
  var Ljtsk = 2 * Math.atan(sinV / (1 + cosV)) / alfa;
  var t = Math.exp(2 / alfa * Math.log((1 + sinU) / cosU / k));
  var pom = (t - 1) / (t + 1);

  do
  {
    var sinB = pom;
    pom = t * Math.exp(e * Math.log((1 + e * sinB) / (1 - e * sinB))); 
    pom = (pom - 1) / (pom + 1);
  } while (Math.abs(pom - sinB) > 1e-15);
  
  var Bjtsk = Math.atan(pom / Math.sqrt(1 - pom * pom));

  /* Pravoúhlé souřadnice ve S-JTSK */  
  a = 6377397.15508;
  f_1 = 299.152812853;
  e2 = 1 - (1 - 1 / f_1) * (1 - 1 / f_1);
  ro = a / Math.sqrt(1 - e2 * Math.sin(Bjtsk) * Math.sin(Bjtsk));
  x = (ro + H) * Math.cos(Bjtsk) * Math.cos(Ljtsk); 
  y = (ro + H) * Math.cos(Bjtsk) * Math.sin(Ljtsk); 
  z = ((1 - e2) * ro + H) * Math.sin(Bjtsk);
  
  /* Pravoúhlé souřadnice v WGS-84 */ 
  var dx = 570.69;
  var dy = 85.69;
  var dz = 462.84; 
  var wz = -5.2611 / 3600 * Math.PI / 180;
  var wy = -1.58676 / 3600 * Math.PI / 180;
  var wx = -4.99821 / 3600 * Math.PI / 180;
  var m = 3.543e-6; 
  xn = dx + (1 + m) * (x + wz * y - wy * z);
  yn = dy + (1 + m) * (-wz * x + y + wx * z);
  zn = dz + (1 + m) * (wy * x - wx * y + z);
  
  /* Geodetické souřadnice v systému WGS-84 */ 
  a = 6378137.0;
  f_1 = 298.257223563;
  a_b = f_1 / (f_1 - 1);
  var p = Math.sqrt(xn * xn + yn * yn);
  var e2 = 1 - (1 - 1 / f_1) * (1 - 1 / f_1);
  var theta = Math.atan(zn * a_b / p);
  var st = Math.sin(theta);
  var ct = Math.cos(theta);
  var t = (zn + e2 * a_b * a * st * st * st) / (p - e2 * a * ct * ct * ct);
  B = Math.atan(t);
  L = 2 * Math.atan(yn / (p + xn));
  H = Math.sqrt(1 + t * t) * (p - a / Math.sqrt(1 + (1 - e2) * t * t));
  
  return (new SMap.Coords.fromWGS(B / Math.PI * 180, L / Math.PI * 180));
}

// aby fungovalo Projection.Krovak
SMap.Coords.prototype.toJTSK = SMap.Coords.prototype.toSJTSK;
SMap.Coords.fromJTSK = SMap.Coords.fromSJTSK;

/**
* Tovární metoda - výroba souřadnic z Puwg1992 (polský souřadnicový systém)
* @author unknown
* @param {int} y Y
* @param {int} x X
* @returns {SMap.Coords}
*/
SMap.Coords.fromPuwg1992 = function(y, x)
{
  var N = (8.3206362430206483E-1 + 1.3481879090606402E-8 * y + 1.5602327664852516E-7 * x - 1.3256276739522121E-14 * y * y + 4.1825950896992127E-15 * x * y - 2.9377381449595037E-16 * x * x - 4.5218702489875804E-22 * y * y * y - 3.9752390006101247E-21 * x * y * y + 6.7313812941904462E-22 * x * x * y - 5.6886908481842923E-23 * x * x * x + 2.2731624246717339E-28 * y * y * y * y - 4.113938338560179E-28 * x * y * y * y - 6.783150621322181E-28 * x * x * y * y + 2.4412235774740029E-28 * x * x * x * y + 5.4276419199397539E-31 * x * x * x * x - 6.4388670061403155E-37 * y * y * y * y * y + 2.0368288420261601E-34 * x * y * y * y * y + 4.393095690025773E-36 * x * x * y * y * y - 2.4158699483191763E-34 * x * x * x * y * y - 1.3676810703780487E-36 * x * x * x * x * y - 6.1662070346142104E-37 * x * x * x * x * x) * 180 / Math.PI;
  var E = (2.1529904669561182E-1 + 2.309998186000915E-7 * y - 1.9968132289432323E-8 * x + 4.8036254777118241E-15 * y * y + 3.906197662448301E-14 * x * y - 4.8275072391845386E-15 * x * x - 2.9467755247811066E-21 * y * y * y + 2.6372320268815168E-21 * x * y * y + 8.9030553688988987E-21 * x * x * y - 8.6938915861748507E-22 * x * x * x - 3.7919111897204254E-28 * y * y * y * y - 1.7774342020473573E-27 * x * y * y * y + 2.2263178567053806E-27 * x * x * y * y + 1.7588476236488152E-27 * x * x * x * y - 3.6555719566702145E-28 * x * x * x * x + 1.4971364431168086E-34 * y * y * y * y * y + 1.0721251670702239E-35 * x * y * y * y * y - 1.464901003129573E-33 * x * x * y * y * y - 2.1018265780491216E-35 * x * x * x * y * y + 7.2158604249699042E-34 * x * x * x * x * y + 2.0298976445575483E-36 * x * x * x * x * x) * 180 / Math.PI;
  return (new SMap.Coords.fromWGS(N, E));
}

/**
* Převede souřadnice do Puwg1992 (polský souřadnicový systém)
* @author unknown
* @returns {Object} puwgCoords Puwg1992 souřadnice
* @returns {number} puwgCoords.x X
* @returns {number} puwgCoords.y Y 
*/
SMap.Coords.prototype.toPuwg1992 = function()
{
  var x = this.lat() * Math.PI / 180;
  var y = this.lon() * Math.PI / 180;
  var cx = -1.6873184841158912E+6 + 7.1740308080610931E+6 * y + 8.4089827999949455E+4 * x - 2.4514780236475319E+6 * y * y - 1.4882979362675548E+6 * x * y + 1.1667953853476048E+6 * x * x + 2.3836063858750649E+6 * y * y * y + 4.7725105592213273E+6 * x * y * y - 2.6700937784006E+6 * x * x * y - 2.0077724548304081E+5 * x * x * x + 1.6544340109570697E+5 * y * y * y * y - 4.7726196774263009E+6 * x * y * y * y - 2.0564038970388323E+6 * x * x * y * y + 1.7055551624134183E+5 * x * x * x * y + 1.9087881651997566E+4 * x * x * x * x - 1.0146215099524893E+5 * y * y * y * y * y + 3.0430736431479454E+3 * x * y * y * y * y + 2.037681378207285E+6 * x * x * y * y * y + 1.0726645753487945E+4 * x * x * x * y * y + 1.8054976024166495E+5 * x * x * x * x * y - 1.7430914509907365E+4 * x * x * x * x * x;
  var cy = -5.2925396175019704E+6 + 3.246150581805408E+5 * y + 6.5982573128608167E+6 * x - 8.4372481974698603E+4 * y * y - 3.9454544085918069E+6 * x * y + 3.7411205534934998E+5 * x * x - 1.2257334379010648E+6 * y * y * y + 5.5874140706162751E+6 * x * y * y + 2.9854418101193309E+6 * x * x * y - 8.7192706190568209E+5 * x * x * x + 9.2642040718702413E+5 * y * y * y * y + 1.1050736306678504E+6 * x * y * y * y - 4.5096171307944506E+6 * x * x * y * y - 3.3980037700653076E+5 * x * x * x * y + 5.1689369445577264E+5 * x * x * x * x - 4.0153157295100391E+2 * y * y * y * y * y - 8.3754242486413941E+5 * x * y * y * y * y + 3.2360315524674952E+3 * x * x * y * y * y + 5.172036182330288E+5 * x * x * x * y * y - 8.9049409680813551E+2 * x * x * x * x * y - 1.1561819681824744E+5 * x * x * x * x * x;
  return {x: cx, y: cy};
}

/**
* zformátuje souřadnice do textového řetězce
* @author N
* @param {string} [type="WGS"] do jakého souřadnicového systému budeme souřadnice formátovat ("WGS", "UTM", "SJTSK")
* @param {string} param pokud jsme zadali "WGS", tak můžeme určit, zda bude formát ve stupních "deg" (defaultní), minutách "min", vteřinách "sec"
* @param {int} precision na kolik desetinných míst zaokrouhlit, defaultně se nastaví číslo závislé na typu projekce; pokud se zadá -1, tak se vytisknou všechna desetinná místa
* @returns {string}
*/
SMap.Coords.prototype.format = function(type, param, precision)
{
  // funkce pro formátování čísla (dle počtu desetinných míst atd.)
  function numFormat(number, digits, digitsWhenNotStripPadding /* počet číslic, kdy necháme na konci nuly */)
  {
    //digitsWhenNotStripPadding = digitsWhenNotStripPadding || 0;
    var result;
    
    if (typeof precision === 'undefined')
      result = number.toFixed(digits);
    else if (precision === -1)
      result = number;
    else
    {
      result = number.toFixed(precision);
      digits = precision;
    }
      
    return ((digitsWhenNotStripPadding == digits) ? result : parseFloat(result));
  }
  
  function latLonLetter(c)
  {
    return {lat: ((c.lat() >= 0) ? "N" : "S"), lon: ((c.lon() >= 0) ? "E" : "W")} 
  }

  // WGS-84
  if (type === "WGS84" || type === "WGS" || typeof type === 'undefined')
  {
    function getDegrees(c) {return (Math.floor(Math.abs(c)));}
    function getMinutes(c) {return ((Math.abs(c) - getDegrees(c)) * 60);}
    function getSeconds(c) {return ((getMinutes(c) - Math.floor(getMinutes(c))) * 60);}

    function formatDegMinSec(c)
    {
      var sec = numFormat(getSeconds(c), 2, 2);
      var min = Math.floor(getMinutes(c));
      var deg = getDegrees(c);
      if (sec == 60) {sec = numFormat(0, 2, 2), min++};
      if (min == 60) {min = Math.floor(0), deg++};
      return deg + "° " + min + "' " + sec + "\"";
    }
    
    function formatDegMin(c)
    {
      var min = numFormat(getMinutes(c), 3, 3);
      var deg = getDegrees(c);
      if (min == 60) {min = numFormat(0, 3, 3), deg++};
      return deg + "° " + min;
    }

    var lat = this.lat();
    var lon = this.lon();
    
    // normalize longitude to -180..180
    lon = (lon + 360 * Math.ceil(Math.abs(lon) / 360)) % 360;
    lon = (lon >= 180) ? (lon - 360) : lon;

    // get seconds
    if (/sec/.test(param))
    {
      lat = latLonLetter(this).lat + " " + formatDegMinSec(lat);
      lon = latLonLetter(this).lon + " " + formatDegMinSec(lon);
    }                                                                                                     
                                                                                                                          
    // get minutes                                                                                                     
    else if (/min/.test(param)/* || typeof type === 'undefined'*/)
    {
      lat = latLonLetter(this).lat + " " + formatDegMin(lat);
      lon = latLonLetter(this).lon + " " + formatDegMin(lon);
    }
                                                                              
    // get degrees
    else //if (/deg/.test(param))
    {
      lat = numFormat(lat, 7);
      lon = numFormat(lon, 7);
    }
    
    return (lat + "  " + lon);        
  }

  // UTM
  else if (type === "UTM")
  {
    var utm = this.toUTM();
    var zone_utm = utm.zone + "U";
    var lon_utm = latLonLetter(this).lon + " " + numFormat(utm.x, 0);
    var lat_utm = latLonLetter(this).lat + " " + numFormat(utm.y, 0);

    return zone_utm + "  " + lon_utm + "  " + lat_utm;  
  }
  
  // S-JTSK
  else if (type === "SJTSK")
  {
    var sjtsk = this.toSJTSK();
    var x = "X " + numFormat(sjtsk.y, 3);
    var y = "Y " + numFormat(sjtsk.x, 3);
    return (y + "  " + x);  
  }

  return "";
}

/**
* načte souřadnice z textového řetězce
* @author N
* @param {string} text řetězec se souřadnicemi
* @returns {SMap.Coords}
*/
SMap.Coords.fromString = function(text)
{
  // změníme desetinnou čárku na tečku
  text = text.replace(/,/g, ".");
  
  // odstraníme všechna slova (žádný formát níže neobsahuje více než jedno písmeno naráz)
  text = text.replace(/[a-záčďéěíňóřšťúůýž]{2,}/gi, "");

  // změníme s.š. na N, atd. (formát Wikipedie: 50°26′12″ s. š., 15°21′12″ v. d.)  
  text = text.replace(/s\.\s*š\./gi, "N");
  text = text.replace(/j\.\s*š\./gi, "S");
  text = text.replace(/z\.\s*d\./gi, "W");
  text = text.replace(/v\.\s*d\./gi, "E");  
  
  var m = text.match(/([+-]?\d+([\.,]\d+)?|N|E|S|W)/gi);
  
  var lat;
  var lon;
  
  if (m === null)
    return null;
  
  // 50.0329  15.834017
  if (m.length == 2 && Math.abs(parseFloat(m[0])) < 90 && Math.abs(parseFloat(m[0])) < 360)
  {
      lat = parseFloat(m[0]);
      lon = parseFloat(m[1]);
  }
  
  // X 1061864.229 Y 643337.950 (SJTSK)
  // Y 643337.950 X 1061864.229 (SJTSK)
  // Y -643337.950 X -1061864.229 (SJTSK)
  // X -1061864.229 Y -643337.950 (SJTSK)
  else if (m.length == 2 && Math.abs(parseFloat(m[0])) > 100000 && Math.abs(parseFloat(m[1])) > 100000)
  {
    var coord1 = Math.abs(parseFloat(m[0]));
    var coord2 = Math.abs(parseFloat(m[1]));
    
    // vždy platí: Y < X
    return SMap.Coords.fromSJTSK((coord1 < coord2) ? coord1 : coord2, (coord1 < coord2) ? coord2 : coord1);
  }
  
  // 33U  E 559730  N 5542622 (UTM, GC.com)  
  else if (m.length == 5
      && (m[3].toUpperCase() === "N" || m[3].toUpperCase() === "S")
      && (m[1].toUpperCase() === "E" || m[1].toUpperCase() === "W"))  
  {
    var coords = SMap.Coords.fromUTM(parseFloat(m[2]), parseFloat(m[4]), parseInt(m[0]), m[3].toUpperCase() === "S");
    lat = coords.lat();
    lon = coords.lon();
  }
  
  // N50.123° E15.45.456°
  else if (m.length == 4 
      && (m[0].toUpperCase() === "N" || m[0].toUpperCase() === "S")
      && (m[2].toUpperCase() === "E" || m[2].toUpperCase() === "W"))
  {
      lat = parseFloat(m[1]);
      lon = parseFloat(m[3]);
      
      if (m[0].toUpperCase() === "S")
          lat *= -1;
      
      if (m[2].toUpperCase() === "W")
          lon *= -1; 
  }
  
  // N50° 01.974  E15° 50.041 (GSAK)
  else if (m.length == 6 
      && (m[0].toUpperCase() === "N" || m[0].toUpperCase() === "S")
      && (m[3].toUpperCase() === "E" || m[3].toUpperCase() === "W"))
  {
      lat = parseFloat(m[1]) + parseFloat(m[2]) / 60;
      lon = parseFloat(m[4]) + parseFloat(m[5]) / 60;
      
      if (m[0].toUpperCase() === "S")
          lat *= -1;
      
      if (m[3].toUpperCase() === "W")
          lon *= -1; 
  }
  
  // N 50° 1' 58.44''  E 15° 50' 2.46'' (GSAK)
  else if (m.length == 8 
      && (m[0].toUpperCase() === "N" || m[0].toUpperCase() === "S")
      && (m[4].toUpperCase() === "E" || m[4].toUpperCase() === "W"))
  {
      lat = parseFloat(m[1]) + parseFloat(m[2]) / 60 + parseFloat(m[3]) / 3600;
      lon = parseFloat(m[5]) + parseFloat(m[6]) / 60 + parseFloat(m[7]) / 3600;
      
      if (m[0].toUpperCase() === "S")
          lat *= -1;
      
      if (m[4].toUpperCase() === "W")
          lon *= -1; 
  }
  
  // 49°52'22.540"N, 15°8'29.821"E (Seznam)
  else if (m.length == 8 
      && (m[3].toUpperCase() === "N" || m[3].toUpperCase() === "S")
      && (m[7].toUpperCase() === "E" || m[7].toUpperCase() === "W"))
  {
      lat = parseFloat(m[0]) + parseFloat(m[1]) / 60 + parseFloat(m[2]) / 3600;
      lon = parseFloat(m[4]) + parseFloat(m[5]) / 60 + parseFloat(m[6]) / 3600;
      
      if (m[3].toUpperCase() === "S")
          lat *= -1;
      
      if (m[7].toUpperCase() === "W")
          lon *= -1; 
  }
  
  // 49°52.540'N, 15°8.821'E (Seznam)
  else if (m.length == 6 
      && (m[2].toUpperCase() === "N" || m[2].toUpperCase() === "S")
      && (m[5].toUpperCase() === "E" || m[5].toUpperCase() === "W"))
  {
      lat = parseFloat(m[0]) + parseFloat(m[1]) / 60;
      lon = parseFloat(m[3]) + parseFloat(m[4]) / 60;
      
      if (m[2].toUpperCase() === "S")
          lat *= -1;
      
      if (m[5].toUpperCase() === "W")
          lon *= -1; 
  }  
  
  // 49.8729278N, 15.1416169E (Seznam)
  else if (m.length == 4 
      && (m[1].toUpperCase() === "N" || m[1].toUpperCase() === "S")
      && (m[3].toUpperCase() === "E" || m[3].toUpperCase() === "W"))
  {
      lat = parseFloat(m[0]);
      lon = parseFloat(m[2]);
      
      if (m[1].toUpperCase() === "S")
          lat *= -1;
      
      if (m[3].toUpperCase() === "W")
          lon *= -1;
  }

  // 50° 15.20' 15° 50.17'
  else if (m.length == 4)
  {
      var latSign = (m[0] >= 0) ? 1 : -1;
      var lonSign = (m[2] >= 0) ? 1 : -1;
      lat = parseFloat(m[0]) + latSign * parseFloat(m[1]) / 60;
      lon = parseFloat(m[2]) + lonSign * parseFloat(m[3]) / 60;
  }  
  
  // 50° 1' 58.44''  15° 50' 2.46'' (např. mavlast.cz)
  else if (m.length == 6)
  {
      var latSign = (m[0] >= 0) ? 1 : -1;
      var lonSign = (m[3] >= 0) ? 1 : -1;  
      lat = parseFloat(m[0]) + latSign * parseFloat(m[1]) / 60 + latSign * parseFloat(m[2]) / 3600;
      lon = parseFloat(m[3]) + lonSign * parseFloat(m[4]) / 60 + lonSign * parseFloat(m[5]) / 3600;
  }  
  
  // pokud jsme dostali nesmysl nebo nepodporovaný formát, tak vrátíme null
  if (!isFinite(lat) || !isFinite(lon) || Math.abs(lat) > 90 || Math.abs(lon) > 180)
    return null;
  
  return (new SMap.Coords.fromWGS(lat, lon));
}

/**
* Tovární metoda - výroba souřadnic ze souřadnicového objektu Google mapy
* @author N
* @param {object} coords souřadnice jako objekt GoogleMaps
* @returns {SMap.Coords}
*/
SMap.Coords.fromGoogle = function(coords)
{
  return SMap.Coords.fromWGS(coords.lat(), coords.lng());
}

/**
* Tovární metoda - výroba souřadnic z události
* @author Seznam
* @param {event} event
* @param {SMap} map
* @returns {SMap.Coords}
*/
SMap.Coords.fromEvent = function(event, map)
{
  return SMap.Pixel.fromEvent(event, map).toCoords(map);
}

/**
* Převod na pixel
* @author Seznam
* @param {SMap} map Mapa, vůči jejímu středu se pozice počítá
* @param {int} [zoom=map.zoom] Zoom
* @returns {SMap.Pixel} Nový pixel
*/
SMap.Coords.prototype.toPixel = function(map, zoom)
{
  var proj = map.getProjection();
  var z = zoom || map.getZoom();
  return proj.coordsToPixel(this, map.getCenter(), z, map.getOrientation());
}

/**
* Vrátí duplikat
* @author Seznam, upraveno by N
* @returns {SMap.Coords} Zduplikovaná souřadnice
*/
SMap.Coords.prototype.clone = function()
{
  return new SMap.Coords(this.lat(), this.lon());
}

/**
* Jsou shodné?
* @author Seznam, upraveno by N
* @param {SMap.Coords} coords
*/
SMap.Coords.prototype.equals = function(coords)
{
  return (this.lat() == coords.lat() && this.lon() == coords.lon());
}

/**
* Ověří, jsou-li tyto souřadnice v průhledu mapy
* @author Seznam
* @param {SMap} map
* @param {bool} [usePadding=false] Má-li se průhled zúžit o paddingy způsobené ovládacími prvky
*/
SMap.Coords.prototype.inMap = function(map, usePadding)
{
  var size = map.getSize();
  var w = Math.round(size.x / 2);
  var h = Math.round(size.y / 2);
  
  var lb = new SMap.Pixel(-w, -h);
  var rt = new SMap.Pixel(w, h);
  
  if (usePadding) {
  	lb.x += map.getPadding("left");
  	lb.y -= map.getPadding("bottom");
  	rt.x -= map.getPadding("right");
  	rt.y += map.getPadding("top");
  }
  
  var px = this.toPixel(map);
  
  return (px.x >= lb.x && px.x <= rt.x && px.y >= lb.y && px.y <= rt.y);
}

/**
* Spočte souřadnici středu mapy tak, aby v novém zoomu zadaný bod zůstal na stejném pixelovém místě
* @author Seznam
* @param {SMap} map Mapa
* @param {int} newZoom Nový zoom
* @returns {SMap.Coords} Nový střed mapy
*/
SMap.Coords.prototype.newCenter = function(map, newZoom)
{
  var px1 = this.toPixel(map);
  var px2 = this.toPixel(map, newZoom);
  var delta = new SMap.Pixel(px2.x-px1.x, px2.y-px1.y);
  return delta.toCoords(map, newZoom);
}

/**
* Převod na řetězec
* @author Seznam, upraveno by N
* @returns {string} Zeserializovaná souřadnice
*/
SMap.Coords.prototype.toString = function()
{
  var coords = this.toPP();
  return "("+coords[0]+","+coords[1]+")";
}

/**
* Zjistí, zda jsou souřadnice ve validním rozsahu (nahrazuje původní funkci)
* @author Seznam, upraveno by N
* @returns {boolean}
*/
SMap.Coords.prototype.isValid = function()
{
  // latitude: using +-90 results in texture errors on poles (so use lower values)
  // longitude:  can be as high/low as we desire (we get rotating world)
  return (this.lat() > -75 && this.lat() < 85 && this.lon() > -360 && this.lon() < 360);
  //return true; 
}

/**
* Zjistí, zda jsou souřadnice ve validním UTM 33 rozsahu
* @author Seznam, N
* @returns {boolean}
*/
SMap.Coords.prototype.isValidUTM33 = function()
{
  return (this.lat() > 38 && this.lat() < 57 && this.lon() > 5 && this.lon() < 25); 
}

// deprecated
SMap.Coords._splitWGS84 = null;
SMap.Coords._fromWGS84str = null;
SMap.Coords.prototype._formatWGS84 = null;

// použito jen v SMap.Route:
SMap.Coords._alphabet = null;
SMap.Coords.stringToCoords = null;
SMap.Coords._parseNumber = null;
SMap.Coords.coordsToString = null;
SMap.Coords._serializeNumber = null;
SMap.Coords.stringToAltitude = null;

// nevyužito
SMap.Coords.prototype.fixedPoint = null;
SMap.Coords.prototype.distanceMiro = null;

//***********************************************************************************************
//***********************************************************************************************
//***********************************************************************************************

SMap.Coords.angDiff = function(angle1, angle2)
{
  var diff = Math.abs(angle1 - angle2);
  diff %= 360;
  
  if (diff > 180)
    diff = 360 - diff;
    
  return diff;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */
/* Vincenty Direct and Inverse Solution of Geodesics on the Ellipsoid (c) Chris Veness 2002-2014  */
/*                                                                                                */
/* from: T Vincenty, "Direct and Inverse Solutions of Geodesics on the Ellipsoid with application */
/*       of nested equations", Survey Review, vol XXIII no 176, 1975                              */
/*       http://www.ngs.noaa.gov/PUBS_LIB/inverse.pdf                                             */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

// Vincenty formula (more precise, uses ellipsoid)
// http://en.wikipedia.org/wiki/Vincenty's_formulae
SMap.Coords.prototype.directVincenty = function(azimuth, distance)
{
  var rad = Math.PI / 180;
  var φ1 = this.lat() * rad
  var λ1 = this.lon() * rad;
  var α1 = azimuth * rad;
  var s = distance;

  var a = 6378137, f = 1 / 298.257223563, b = a * (1 - f);

  var sinα1 = Math.sin(α1);
  var cosα1 = Math.cos(α1);

  var tanβ1 = (1-f) * Math.tan(φ1), cosβ1 = 1 / Math.sqrt((1 + tanβ1 * tanβ1)), sinβ1 = tanβ1 * cosβ1;
  var σ1 = Math.atan2(tanβ1, cosα1);
  var sinα = cosβ1 * sinα1;
  var cosSqα = 1 - sinα * sinα;

  var u2 = 1/4 * cosSqα * ((a * a) / (b * b) - 1);
  var B0 = (1 + u2 * (1/64 * u2 * (u2 * (80 - 175 * u2) - 48) + 1));
  var B2 = u2 * (u2 * (5/8 * u2 * (7 * u2 - 3) + 1) - 1);
  var B4 = 1/32 * u2 * u2 * (u2 * (-35 * u2 + 12) - 4);
  var B6 = 1/24 * u2 * u2 * u2 * (5 * u2 - 1);
  var B8 = -5/256 * u2 * u2 * u2 * u2;
  
  var σ = s / (b * B0), σʹ, iterations = 0;
  
  do {
      var cos2σM = Math.cos(2 * σ1 + σ);
      var sinσ = Math.sin(σ);
      var cosσ = Math.cos(σ);
      
      σʹ = σ;

      σ = (s / b - sinσ * ( 
        + B2 * cos2σM 
        + cosσ * (B4 * (4 * cos2σM * cos2σM - 2)
        + B8 * (8 * cosσ * cosσ - 4) * (8 * cos2σM * cos2σM * (cos2σM * cos2σM - 1) + 1)) 
        + B6 * (4 * cosσ * cosσ - 1) * cos2σM * (4 * cos2σM * cos2σM - 3))) / B0;                   

  } while (Math.abs(σ - σʹ) > 1e-15 && ++iterations < 2000);

  // will fail to converge when choosing huge distances like 200 000 km
  //if (iterations >= 200)
  //  throw new Error('Formula failed to converge'); // not possible?

  // get φ2
  var x = sinβ1 * sinσ - cosβ1 * cosσ * cosα1;
  var φ2 = Math.atan2(sinβ1 * cosσ + cosβ1 * sinσ * cosα1, (1 - f) * Math.sqrt(sinα * sinα + x * x)) / rad;

  // get λ2
  var λ = Math.atan2(sinσ * sinα1, cosβ1 * cosσ - sinβ1 * sinσ * cosα1);
  
  var λ12 = λ - f * sinα * (σ + 1/768 * f * cosSqα * (
    192 * (1 + f * (1 + f)) * (-σ + sinσ * cos2σM)
  + f * cosSqα * (
  (48 + 108 * f) * (3 * σ - sinσ * (2 * cos2σM * (2 - cosσ * cos2σM) + cosσ))
  + f * cosSqα *(-150 * σ  + sinσ * ((225 * cos2σM - cosσ * (180 * cos2σM * cos2σM - 90)) 
  + (4 * cosσ * cosσ - 1) * (cos2σM * (20 * cos2σM * cos2σM - 15)))))));       

  var λ2 = ((λ1 + λ12) / rad + 540) % (360) - 180;  // normalise to -180...+180

  // get α2
  //var α2 = Math.atan2(sinα, -x) / rad;
  //α2 = (α2 + 360) % 360; // normalise to 0...360
  
  return SMap.Coords.fromWGS(φ2, λ2);
  //return {lat2: φ2, lon2: λ2, finalBearing: α2};
}

SMap.Coords.prototype.inverseVincenty = function(point)
{
  var rad = Math.PI / 180;
  var φ1 = this.lat() * rad, λ1 = this.lon() * rad;
  var φ2 = point.lat() * rad, λ2 = point.lon() * rad;
  
  var a = 6378137.0, f = 1 / 298.257223563, b = a * (1 - f);
  
  var λ12 = λ2 - λ1;
  var tanβ1 = (1 - f) * Math.tan(φ1), cosβ1 = 1 / Math.sqrt((1 + tanβ1 * tanβ1)), sinβ1 = tanβ1 * cosβ1;
  var tanβ2 = (1 - f) * Math.tan(φ2), cosβ2 = 1 / Math.sqrt((1 + tanβ2 * tanβ2)), sinβ2 = tanβ2 * cosβ2;

  var λ = λ12, λʹ, iterations = 0;
  
  do {
      // deal with upper and lower half of antipodal region (from Rapp, Vol. II, page 50)
      if (iterations > 20)
      {
        if (typeof previousλ === 'undefined') var previousλ = new Array(3);
        if (isFinite(previousλ[0])) λ = (2 * λ + 3 * previousλ[1] + previousλ[0]) / 6;
        previousλ[0] = previousλ[1], previousλ[1] = λ;
      }

      var sinλ = Math.sin(λ), cosλ = Math.cos(λ);
      var sinSqσ = (cosβ2 * sinλ) * (cosβ2 * sinλ) + (cosβ1 * sinβ2 - sinβ1 * cosβ2 * cosλ) * (cosβ1 * sinβ2 - sinβ1 * cosβ2 * cosλ);
      var sinσ = Math.sqrt(sinSqσ);
      if (sinσ == 0) return {distance: 0, initialBearing: 0, finalBearing: 180};  // co-incident points
      var cosσ = sinβ1 * sinβ2 + cosβ1 * cosβ2 * cosλ;
      var σ = Math.atan2(sinσ, cosσ);
      var sinα = cosβ1 * cosβ2 * sinλ / sinσ;
      var cosSqα = 1 - sinα * sinα;
      var cos2σM = cosσ - 2 * sinβ1 * sinβ2 / cosSqα;
      if (isNaN(cos2σM)) cos2σM = 0;  // equatorial line: cosSqα=0 (§6)
      λʹ = λ;
 
      λ = λ12 + f * sinα * (σ + 1/768 * f * cosSqα * (
        192 * (1 + f * (1 + f)) * (-σ + sinσ * cos2σM)
      + f * cosSqα * (
      (48 + 108 * f) * (3 * σ - sinσ * (2 * cos2σM * (2 - cosσ * cos2σM) + cosσ))
      + f * cosSqα *(-150 * σ  + sinσ * ((225 * cos2σM - cosσ * (180 * cos2σM * cos2σM - 90)) 
      + (4 * cosσ * cosσ - 1) * (cos2σM * (20 * cos2σM * cos2σM - 15)))))));

      //log (λ - λʹ);
  
  } while (Math.abs(λ - λʹ) > 1e-15 && ++iterations < 200);
     
  // deal with the antipodal region
  if (iterations == 200)
  {
    iterations = 0;
    var λʹ = 0;  
    var λ12ʹ = (λ12 > 0) ? (Math.PI - λ12) : (-Math.PI - λ12);
    var cosSqα = 1/2;
    var cos2σM = 0;
    var σ = Math.PI - Math.abs(Math.atan2(sinβ1, cosβ1) + Math.atan2(sinβ2, cosβ2));
    var sinσ = Math.sin(σ), cosσ = Math.cos(σ);
    var sinα;
    var lastSinα, lastσ, lastλʹ;
    var sameSigns = 0, lastSign = 1;

    do {
        lastSinα = sinα, lastσ = σ, lastλʹ = λʹ;

        sinα = (λ12ʹ - λʹ) / (f * (σ + 1/768 * f * cosSqα * (
          192 * (1 + f * (1 + f)) * (-σ + sinσ * cos2σM)
        + f * cosSqα * (
        (48 + 108 * f) * (3 * σ - sinσ * (2 * cos2σM * (2 - cosσ * cos2σM) + cosσ))
        + f * cosSqα *(-150 * σ  + sinσ * ((225 * cos2σM - cosσ * (180 * cos2σM * cos2σM - 90)) 
        + (4 * cosσ * cosσ - 1) * (cos2σM * (20 * cos2σM * cos2σM - 15))))))))

        cosSqα = 1 - sinα * sinα;       

        var sinλʹ = sinα * sinσ / (cosβ1 * cosβ2);
        var cosλʹ = Math.sqrt(1 - sinλʹ * sinλʹ) * (((cosσ - sinβ1 * sinβ2) / (cosβ1 * cosβ2)) > 0 ? -1 : 1)
        λʹ = Math.atan2(sinλʹ, cosλʹ);

        // deal with suckers in far-left/rigth antipodal region, that would need millions of iterations
        // -- we speed up change of λ significantly to overcome slowness of change
        var diffλ = λʹ - lastλʹ;

        if (iterations > 100 && diffλ / lastSign >= 0)
        {
           lastSign = diffλ > 0 ? 1 : -1;
           λʹ += Math.min(0.1, Math.pow(3, sameSigns++) * diffλ);
           sinλʹ = Math.sin(λʹ); cosλʹ = Math.cos(λʹ);
        }
        else
        {
          sameSigns = 0;
          lastSign = diffλ > 0 ? 1 : -1;
        }

        sinσ = Math.sqrt((cosβ2 * sinλʹ) * (cosβ2 * sinλʹ) + (cosβ1 * sinβ2 + sinβ1 * cosβ2 * cosλʹ) * (cosβ1 * sinβ2 + sinβ1 * cosβ2 * cosλʹ));
        cosσ = (sinβ1 * sinβ2 - cosβ1 * cosβ2 * cosλʹ);
        σ = Math.atan2(sinσ, cosσ);
        cos2σM = cosσ - 2 * sinβ1 * sinβ2 / cosSqα;

        //log(λʹ - lastλʹ);
     
       /* track all we can to prevent sloooowly converging coords to leave the chains */
    } while ((Math.abs(λʹ - lastλʹ) > 1e-15 || Math.abs(sinα - lastSinα) > 1e-15   
        || Math.abs(σ - lastσ) > 1e-15) && ++iterations < 1000);
              
    var sinα1 = sinα / cosβ1;
    var cosα1 = Math.sqrt(1 - sinα1 * sinα1) * ((cosβ1 * sinβ2 + sinβ1 * cosβ2 * cosλʹ) < 0 ? -1 : 1);
  }

  var u2 = 1/4 * cosSqα * ((a * a) / (b * b) - 1);

  var s = b * (σ 
  + u2 * ((σ - sinσ * cos2σM)
  + 1/192 * u2 * (48 * (-3 * σ + 4 * sinσ * cos2σM - sinσ * cosσ * (2 * cos2σM * cos2σM - 1))
  + u2 * (8 * (30 * σ  - sinσ * (45 * cos2σM - 18 * cosσ * (2 * cos2σM * cos2σM - 1) + (4 * cosσ * cosσ - 1) * cos2σM * (4 * cos2σM * cos2σM - 3)))
  + u2 * (-525 * σ + sinσ * (840 * cos2σM - 420 * cosσ * (2 * cos2σM * cos2σM - 1) 
    + 40 * (4 * cosσ * cosσ - 1) * cos2σM * (4 * cos2σM * cos2σM - 3) 
    - 15 * cosσ * (2 * cosσ * cosσ - 1) * (8 * cos2σM * cos2σM * (cos2σM * cos2σM - 1) + 1)))
  ))));
  
  // get azimuth
  var α1 = Math.atan2(cosβ2 * sinλ,  cosβ1 * sinβ2 - sinβ1 * cosβ2 * cosλ) / rad;
  var α2 = Math.atan2(cosβ1 * sinλ, -sinβ1 * cosβ2 + cosβ1 * sinβ2 * cosλ) / rad;
  
  if (typeof cosα1 !== 'undefined')   /* antipodal */
  {
    α1 = Math.atan2(sinα1, cosα1) / rad;
    α2 = Math.atan2(sinα, -sinβ1 * sinσ + cosβ1 * cosσ * cosα1) / rad;
  }  
  
  // normalise to 0...360
  α1 = (α1 + 360) % 360;
  α2 = (α2 + 360) % 360;

  if (!isFinite(s)) return {distance: 1e6, initialBearing: 1e6, finalBearing: 1e6};

  return {distance: s, initialBearing: α1, finalBearing: α2};
}

// http://en.wikipedia.org/wiki/Geographical_distance
// http://en.wikipedia.org/wiki/Great-circle_distance
// http://en.wikipedia.org/wiki/Geodesics_on_an_ellipsoid
// issus:
//   - doesn't handle correctly antipodal or near antipodal points, 
//   - 0 0, 0 180 computes over equator, but should go through the pole
//   - distanceMiro is better for distance < 35 m, but otherwise worse
SMap.Coords.prototype.distanceLambert = function(target)
{
    var a = 6378137, f = 1 / 298.257223563;
    
    var rad = Math.PI / 180;
    var φ1 = this.lat() * rad;
    var λ1 = this.lon() * rad;
    var φ2 = target.lat() * rad;
    var λ2 = target.lon() * rad;
    
    var cosλ12 = Math.cos(λ2 - λ1);
    
    var sinβ1 = (1 - f) * Math.sin(φ1), cosβ1 = Math.cos(φ1);
    var t = Math.sqrt(sinβ1 * sinβ1 + cosβ1 * cosβ1);
    sinβ1 /= t, cosβ1 /= t;    

    var sinβ2 = (1 - f) * Math.sin(φ2), cosβ2 = Math.cos(φ2);
    var t = Math.sqrt(sinβ2 * sinβ2 + cosβ2 * cosβ2);
    sinβ2 /= t, cosβ2 /= t;    

    var cosσ = sinβ1 * sinβ2 + cosβ1 * cosβ2 * cosλ12;
    var σ = Math.acos(cosσ);

    // very short distances (< 0.15 m)
    if (cosσ >= 1 - 1E-13)
    {
      //σ = Math.atan2(Math.sqrt(Math.pow(Math.cos(φ2) * Math.sin(λ2 - λ1), 2) + Math.pow(Math.cos(φ1) * Math.sin(φ2) - Math.sin(φ1) * Math.cos(φ2) * cosλ12, 2)),
      //  Math.sin(φ1) * Math.sin(φ2) + Math.cos(φ1) * Math.cos(φ2) * cosλ12);

      var S1 = Math.sin((φ2 - φ1) / 2);
      var S2 = Math.sin((λ2 - λ1) / 2);  
      σ = 2 * Math.asin(Math.sqrt(S1 * S1 + Math.cos(φ1) * Math.cos(φ2) * S2 * S2)) 
      cosσ = Math.cos(σ);
    }

    var sinσ = Math.sin(σ);
    
    // handle φ1 = 90 && φ2 = -90 (correctly) and also prevent antipodal error (but gives still incorrect result)
    if (cosσ <= -1)
      cosσ = -1 + 1E-13;
    // handle other possible problems (?)
    else if (cosσ >= 1)
      cosσ = 1 - 1E-13;

    var XY = ((σ + sinσ * cosσ) * (sinβ1 * sinβ1 + sinβ2 * sinβ2) - (sinσ + σ * cosσ) * (2 * sinβ1 * sinβ2)) / (1 - cosσ * cosσ);
    var d = a * (σ - f / 2 * XY);

    //log([σ, d, XY]);
    return d;
};

// initial bearing - haversine version
// average error: 0.07°, max error (excluding antipodal region): 0.26°
// -- suffers from antipodal problems
SMap.Coords.prototype.initialBearingHaversine = function(target)
{
  var rad = Math.PI / 180;
  var φ1 = this.lat() * rad, λ1 = this.lon() * rad;
  var φ2 = target.lat() * rad, λ2 = target.lon() * rad;

  var y = Math.sin(λ2 - λ1);
  var x = Math.cos(φ1) * Math.tan(φ2) - Math.sin(φ1) * Math.cos(λ2-λ1);
  return (Math.atan2(y, x) / rad + 360) % 360;
}

SMap.Coords.prototype.finalBearingHaversine = function(target)
{
  return target.initialBearingHaversine(this);
}

// haversine version of direct geodetic problem,
// - takes 60% of time spent by direct Vincenty function
// - maximum error around 5 meters per 1 km, average error around 3 m/km
// - source: http://www.movable-type.co.uk/scripts/latlong.html
SMap.Coords.prototype.directHaversine = function(azimuth, distance)
{
    var rad = Math.PI / 180;
    var φ1 = this.lat() * rad;
    var λ1 = this.lon() * rad;
    var α1 = azimuth * rad;

    var R = 6371000 /* mean earth radius */;
    
    var cosdR = Math.cos(distance / R), sindR = Math.sin(distance / R);
    var sinφ1 = Math.sin(φ1), cosφ1 = Math.cos(φ1);

    var φ2 = Math.asin(sinφ1 * cosdR + cosφ1 * sindR * Math.cos(α1));
    var λ2 = λ1 + Math.atan2(Math.sin(α1) * sindR * cosφ1, cosdR - sinφ1 * Math.sin(φ2));

    return SMap.Coords.fromWGS(φ2 / rad, λ2 / rad);
};

// distance through earth
// from: http://en.wikipedia.org/wiki/Geographical_distance
SMap.Coords.prototype.tunnelDistance = function(target)
{
  var rad = Math.PI / 180;
  var φ1 = this.lat() * rad, λ1 = this.lon() * rad;
  var φ2 = target.lat() * rad, λ2 = target.lon() * rad;
  var R = 6371000;
    
  var ΔX = Math.cos(φ2) * Math.cos(λ2) - Math.cos(φ1) * Math.cos(λ1);
  var ΔY = Math.cos(φ2) * Math.sin(λ2) - Math.cos(φ1) * Math.sin(λ1);
  var ΔZ = Math.sin(φ2) - Math.sin(φ1);
  var Ch = Math.sqrt(ΔX * ΔX + ΔY * ΔY + ΔZ * ΔZ);
  
  return Ch * R;
}

// closest point on a great circle path
// p1, p2 --> line to find the nearest point on; this --> point from where we find nearest point
// from: http://www.movable-type.co.uk/scripts/latlong.html (Cross-track distance)
SMap.Coords.prototype.nearestPoint = function(p1, p2)
{
    var rad = Math.PI / 180;
    var R = 6371000;

    var d13 = p1.distance(this);
    var θ12 = p1.azimuth(p2) * rad, θ13 = p1.azimuth(this) * rad;
    
    var dXt = Math.asin(Math.sin(d13 / R) * Math.sin(θ13 - θ12)) * R;
    var dAt = Math.acos(Math.cos(d13 / R) / Math.cos(dXt / R)) * R;

    // když je bod "před" prvním bodem přímky, je třeba úhel upravit o 180°
    if (SMap.Coords.angDiff(p1.azimuth(this), p1.azimuth(p2)) >= 90)
      θ12 += 180 * rad;

    return p1.projectPoint(θ12 / rad, dAt);
};

//--------------------------------

SMap.Coords.prototype.projectPoint = function(azimuth, distance, options)
{
  // somewhat quicker version, error 5 meters per 1 km, unsuitable for very long distances
  if (options && options.quick && distance < 1000 * 1000)
    return this.directHaversine(azimuth, distance);
  else
    return this.directVincenty(azimuth, distance);
}

SMap.Coords.prototype.distance = function(target, options)
{
  // Lambertova formule (rychlejší, méně přesná)
  if (options && options.quick)
  {
    var distance = this.distanceLambert(target);
    
    // use Vincenty if possibly antipodal
    if (distance < 18000 * 1000)
      return dist;
  }
  
  // Vincentova formule
  return this.inverseVincenty(target).distance;
}

SMap.Coords.prototype.azimuth = function(target, options)
{
  // haversine version (quick, but less accurate)
  if (options && options.quick)
  {
    var azimuth = this.initialBearingHaversine(target);
    
    // prevent usage on possible antipodal coords
    if (Math.abs(this.lon() - target.lon()) < 170)
      return azimuth;
  }
  
  return this.inverseVincenty(target).initialBearing;
}

SMap.Coords.prototype.initialBearing = SMap.Coords.prototype.azimuth;

SMap.Coords.prototype.finalBearing = function(target, options)
{
  return this.inverseVincenty(target).finalBearing;
}

SMap.Coords.prototype.midpoint = function(target, ratio)
{
  var result = this.inverseVincenty(target);
  ratio = isFinite(ratio) ? ratio : 1/2;
  return this.projectPoint(result.initialBearing + (ratio < 0 ? 180 : 0), result.distance * Math.abs(ratio));
}

//----------------------------------------------

// následující funkce slouží hlavně k vykreslování šipek na přímkách, kde nechceme skutečný azimut
// -- funkce dávají jen přibližné hodnoty (zaokrouhlené na 1 pixel v aktuálním zoomu)
// -- nepoužívat pro nic důležitého!!!!
SMap.Coords.prototype.pixelProject = function(azimuth, pixelDistance)
{
  if (pixelDistance == 0)
    return this;

  azimuth = ((azimuth + 360) % 360) * Math.PI / 180;
  var p1 = this.toPixel(m, m.getZoom());
  
  var x = p1.x + Math.cos(azimuth) * pixelDistance;
  var y = p1.y + Math.sin(azimuth) * pixelDistance;

  var newPoint = new SMap.Pixel(x, y);
  return newPoint.toCoords(m, m.getZoom());
}

SMap.Coords.prototype.pixelDistance = function(endPoint)
{
  var p1 = this.toPixel(m, m.getZoom());
  var p2 = endPoint.toPixel(m, m.getZoom());
  return p1.distance(p2);
}

SMap.Coords.prototype.pixelAzimuth = function(endPoint)
{
  var p1 = this.toPixel(m, m.getZoom());
  var p2 = endPoint.toPixel(m, m.getZoom());
  
  var azimuth = Math.atan2(p2.y - p1.y, p2.x - p1.x) * 180 / Math.PI;
  return ((azimuth + 360) % 360);
}

window.mapyCZCoords_Loaded = true;